import datetime as dt
import inspect
import json
import re
import shlex
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Optional, get_args, get_origin

from .exceptions import ParamNotFoundError, PosParamsCountError


def validate_type(data_type: Any, value: str):
    if data_type is Any:
        return value
    elif data_type is str:
        return str(value)
    elif data_type is int:
        return int(value)
    elif data_type is float:
        return float(value)
    # elif data_type is bytes:
    #     return bytes(value)
    elif data_type is dt.date:
        return dt.date.fromisoformat(value)
    elif data_type is dt.datetime:
        return dt.datetime.fromisoformat(value)
    elif data_type is Path:
        p = Path(value)
        if not p.exists():
            raise FileNotFoundError(f'File not found: "{value}"')
        return p
    elif data_type is dict:
        return json.loads(value)
    elif data_type is list or get_origin(data_type) is list:
        list_type = get_args(data_type)
        return [validate_type(list_type[0] if list_type else str,
                              x) for x in value.split(' ')]
    else:
        raise NotImplementedError(f'No parser implemented for data type "{data_type}"')


@dataclass
class CommandOption:
    default: Any
    help: Optional[str] = None
    keyword_args: tuple[str, ...] = field(default_factory=tuple)

    name: Optional[str] = field(init=False, default=None)
    data_type: Any = field(init=False, default=Any)

    @property
    def is_required(self):
        return self.default is ...

    @property
    def is_positional_arg(self):
        return len(self.keyword_args) == 0

    def validate(self, value: Any) -> Any:
        return validate_type(self.data_type, value)


def Option(
        default: Any,
        *keyword_args: str,
        help: str = None
):
    return CommandOption(
        default=default,
        help=help,
        keyword_args=keyword_args,
    )


def get_cmd_args(cmd: str) -> tuple[list[str], dict[str, str]]:
    positional_args = []
    keyword_params = {}

    is_positional_arg = True
    skip_position = None

    cmd_split = shlex.split(cmd)
    for i, _arg in enumerate(cmd_split):
        if skip_position is not None and i <= skip_position:
            continue

        if _arg.startswith('-'):
            is_positional_arg = False

        if is_positional_arg:
            positional_args.append(_arg)
        else:
            keyword_params[_arg] = (
                cmd_split[i + 1] if i + 1 < len(cmd_split)
                # In case of "store_true"
                else True
            )
            skip_position = i + 1

    return positional_args, keyword_params


def keyword_arg_to_name(keyword_arg: str) -> str:
    """ Formats a string from '--quiet-v2' to 'quiet_v2' """
    return re.sub('^-+', '', keyword_arg).replace('-', '_')


def get_default_args(func):
    signature = inspect.signature(func)
    return [v.default if v is not inspect.Parameter.empty else None
            for v in signature.parameters.values()]


def parse_input_args(args: tuple[Any, ...], commands: set[str]) -> tuple[
    Optional[str], list[str], list[str]
]:
    """
    Extracts the:
     - global options
     - command
     - command options
     from the passed list or arguments
    """
    global_options, cmd_options, cmd = [], [], None
    for arg in args:
        if cmd is None and arg in commands:
            cmd = arg
            continue

        if cmd is None:
            global_options.append(arg)
        else:
            cmd_options.append(arg)
    return cmd, global_options, cmd_options


def convert_args_to_dict(input_args: list[str],
                         options: list[CommandOption]) -> dict:
    _input_pos_args, _input_keyword_args = get_cmd_args(' '.join(input_args))

    positional_args, keyword_args = [], {}
    for _arg in options:
        if _arg.is_positional_arg:
            positional_args.append(_arg)
        for _keyword_arg in _arg.keyword_args:
            keyword_args[_keyword_arg] = _arg.name or keyword_arg_to_name(sorted(_arg.keyword_args)[0])

    # Positional arguments
    if len(_input_pos_args) != len(positional_args):
        raise PosParamsCountError(
            f'Expected {len(positional_args)} positional values but got {len(_input_pos_args)}',
            expected_count=len(positional_args),
            count=len(_input_pos_args)
        )
    fn_args = {
        _param.name: _param.validate(_param_value)
        for _param, _param_value in zip(positional_args, _input_pos_args)
    }

    # Keyword Arguments
    for _keyword_arg_key, _keyword_arg_value in _input_keyword_args.items():
        _keyword_param = keyword_args.get(_keyword_arg_key)
        if not _keyword_param:
            raise ParamNotFoundError(f'Could not find param {_keyword_arg_key}',
                                     _keyword_arg_key)
        fn_args[_keyword_param] = _keyword_arg_value

    # We fill optional fields with None
    for _arg in options:
        _arg_name = _arg.name or keyword_arg_to_name(sorted(_arg.keyword_args)[0])
        if _arg_name not in fn_args:
            fn_args[_arg_name] = None

    return fn_args



import sys
from dataclasses import dataclass, field
from typing import Optional, Any

from .command import CommandGroup, ShowHelpError
from .exceptions import (
    CommandNotFoundError,
    ParamNotFoundError, PosParamsCountError
)
from .formatter import Formatter, RichFormatter


@dataclass
class Cli:
    description: Optional[str] = None
    formatter: Formatter = field(default_factory=RichFormatter)

    _group: CommandGroup = field(init=False, default_factory=CommandGroup)

    def __post_init__(self):
        self._group.help = self.description

    @property
    def commands(self):
        return self._group.commands

    def run(self):
        try:
            _, *args = sys.argv
        except ValueError:
            return
        self.run_with_args(*args)

    def run_with_args(self, *args):
        try:
            return self._group.run_with_args(*args)
        except CommandNotFoundError as e:
            raise e
        except ShowHelpError as e:
            self.formatter.print_help(group=e.group,
                                      command=e.command,
                                      parent_args=e.parent_args)
        except ParamNotFoundError as e:
            self.formatter.print_param_error(e.key)
            return
        except PosParamsCountError as e:
            self.formatter.print_count_error(e.expected_count, e.count)
            return

    def command(self, cmd: str, help: str = None):
        return self._group.command(cmd=cmd, help=help)

    def add_option(self, default: None, *args, help: str = None, data_type: Any = bool):
        self._group.add_option(default, *args, help=help, data_type=data_type)

    def add_sub_parser(self, cmd: str, description: Optional[str] = None) -> CommandGroup:
        cmd_group = CommandGroup(name=cmd, help=description)
        self._group.add_group(cmd_group)
        return cmd_group



import sys
from dataclasses import dataclass, field
from typing import Optional

from rich.console import Console
from rich.table import Table, Padding

from .base import Formatter, Titles, get_options_str
from ..command import Command, CommandOption, ParentArgs, CommandGroup


def pad(s: str, padding_left: int = 1):
    return Padding(s, (0, padding_left))


def fmt_option(option: CommandOption,
               show_full: bool = False,
               color: str = 'white') -> str:
    if option.is_positional_arg:
        return f'[{color}]<{option.name}>[/{color}]'
    elif show_full:
        first_arg, *other_args = option.keyword_args
        if other_args:
            other_args = ', '.join(other_args)
            return f'[{color}]{first_arg}[/{color}] ({other_args})'
        else:
            return first_arg
    else:
        return '[' + sorted(option.keyword_args)[-1] + ']'


def fmt_cmd_options(options: list[CommandOption]) -> str:
    return (' '.join([fmt_option(x) for x in options])
            if options else ''  # '[<arg1>] ... [<argN>]'
            )


def get_usage(global_options: list[CommandOption],
              command: str = None,
              command_options: list[CommandOption] = None,
              parent_args: ParentArgs = None):
    parent_args = parent_args or []
    _global_options = ' '.join(['[' + sorted(x.keyword_args)[-1] + ']' for x in global_options])
    command = f'[underline]{command}[/underline]' if command else '<command>'
    cmds = [sys.argv[0].split('/')[-1]] + [x[0] for x in parent_args]
    cmds = ' '.join(f'[underline]{x}[/underline]' for x in cmds)

    usage = cmds
    if _global_options:
        usage = f'{usage} {_global_options}'
    usage = f'{usage} {command}'
    if command_options:
        usage = f'{usage} {fmt_cmd_options(command_options)}'

    return usage


@dataclass(frozen=True)
class RichTitles(Titles):
    GLOBAL_OPTIONS = f'[bold bright_white]{Titles.GLOBAL_OPTIONS}[/bold bright_white]'
    AVAILABLE_CMDS = f'[bold bright_white]{Titles.AVAILABLE_CMDS}[/bold bright_white]'
    COMMANDS = f'[bold bright_white]{Titles.COMMANDS}[/bold bright_white]'
    USAGE = f'[bold bright_white]{Titles.USAGE}[/bold bright_white]'
    DESCRIPTION = f'[bold bright_white]{Titles.DESCRIPTION}[/bold bright_white]'
    ARGUMENTS = f'[bold bright_white]{Titles.ARGUMENTS}[/bold bright_white]'
    OPTIONS = f'[bold bright_white]{Titles.OPTIONS}[/bold bright_white]'


@dataclass
class RichFormatter(Formatter):
    _console: Console = field(init=False,
                              default_factory=lambda: Console(markup=True, highlight=False))
    cmd_color: str = 'cyan'
    option_color: str = 'cyan'

    def _color_opt(self, opt: str):
        return f'[{self.option_color}]{opt}[/{self.option_color}]'

    def _color_cmd(self, cmd: str):
        return f'[{self.cmd_color}]{cmd}[/{self.cmd_color}]'

    def __post_init__(self):
        self.print_fn = self._console.print

    def _print_options(self, options: list[CommandOption]):
        self.print_rows([(self._color_opt(_name or '') + _other_args, _help)
                         for _name, _other_args, _help in get_options_str(options)])

    def print_rows(self, rows: list[tuple[str, Optional[str]]]):
        table = Table(show_header=False, box=None, padding=(0, self.col_space))
        table.add_column(width=self.col_size)
        table.add_column()
        for row in rows:
            table.add_row(*row)
        self.print_fn(table)

    def print_cli_help(self,
                       group: CommandGroup):
        self.print_fn(RichTitles.USAGE, '\n', get_usage(group.options), '\n')
        if group.options:
            self.print_fn(RichTitles.GLOBAL_OPTIONS)
            self._print_options(group.options)
            print()

        self.print_fn(RichTitles.AVAILABLE_CMDS)
        self.print_rows([(f' {self._color_cmd(_command.name or "")}', _command.help) for _command in
                         group.commands.values()])

        if group.help:
            self.print_fn(f"\n{RichTitles.DESCRIPTION}\n {group.help}\n")

    def print_cmd_help(self,
                       command: Command,
                       options: list[CommandOption],
                       parent_args: ParentArgs = None):
        usage = get_usage(
            global_options=options,
            command=command.name,
            command_options=command.options,
            parent_args=parent_args
        )
        self.print_fn(RichTitles.USAGE, '\n', usage, '\n')

        if command.positional_args:
            self.print_fn(RichTitles.ARGUMENTS)
            self.print_rows(
                [(fmt_option(option, color=self.option_color), option.help)
                 for option in command.positional_args])
        if command.keyword_args:
            self.print_fn('\n' + RichTitles.OPTIONS)
            self._print_options(command.keyword_args)

        global_options = options + [parent_option for parent_arg in (parent_args or [])
                                    for parent_option in parent_arg[1]]
        if global_options:
            self.print_fn('\n' + RichTitles.GLOBAL_OPTIONS)
            self._print_options(global_options)
        if command.help:
            self.print_fn(f'\n{RichTitles.DESCRIPTION}\n {command.help}\n')

    def print_cmd_group_help(self,
                             group: CommandGroup,
                             parent_args: ParentArgs):

        parent_commands = [sys.argv[0].split('/')[-1]] + [x[0] for x in parent_args]
        commands_str = []
        for i, (cmd_name, cmd) in enumerate(group.commands.items()):
            _cmds = ''.join(
                f'[underline]{x}[/underline] ' + (
                    f'{fmt_cmd_options(group.options)} ' if cmd_lvl == len(parent_commands) - 1 else '')
                for cmd_lvl, x in enumerate(parent_commands + [cmd_name]))

            _line = f'{"" if i == 0 else "or: ":>5}{_cmds}{fmt_cmd_options(cmd.options)}'
            commands_str.append(_line)
        commands_str = '\n'.join(commands_str)

        self.print_fn(RichTitles.USAGE)
        self.print_fn(commands_str)

        print()

        self.print_fn(RichTitles.COMMANDS)
        for cmd_name, cmd in group.commands.items():
            self.print_fn(pad(f'[underline]{cmd_name}[/underline]', padding_left=2))
            if cmd.help:
                self.print_fn(pad(cmd.help, padding_left=4))
                print()
            if cmd.options:
                self.print_rows([(fmt_option(opt, show_full=True, color=self.option_color), opt.help)
                                 for opt in cmd.options])
                print()

        if group.options:
            self.print_fn(RichTitles.OPTIONS)
            self._print_options(group.options)
            print()

        global_options = [parent_option
                          for parent_arg in (parent_args or [])
                          for parent_option in parent_arg[1]]
        if global_options:
            self.print_fn(RichTitles.GLOBAL_OPTIONS)
            self._print_options(global_options)
            print()

        if group.help:
            self.print_fn(RichTitles.DESCRIPTION)
            self.print_fn(pad(group.help))

    def print_cmd_error(self, cmd: str):
        self.print_fn(f'[red]Unknown command "[bold]{cmd}[/bold]"[/red]')

    def print_param_error(self, key: str) -> None:
        self.print_fn(f"[red]Could not find value for [bold]{key!r}[/bold][/red]")

    def print_count_error(self, expected_count: int, count: int):
        self.print_fn(f'[red]Expected {expected_count} positional arguments but got {count}[/red]')



from piou import Cli, Option

cli = Cli(description='A CLI tool')

cli.add_option(None, '-q', '--quiet', help='Do not output any message')
cli.add_option(None, '--verbose', help='Increase verbosity')


@cli.command(cmd='foo',
             help='Run foo command')
def foo_main(
        quiet: bool,
        verbose: bool,
        foo1: int = Option(..., help='Foo arguments'),
        foo2: str = Option(..., '-f', '--foo2', help='Foo2 arguments'),
        foo3: str = Option(None, '-g', '--foo3', help='Foo3 arguments'),
):
    pass


@cli.command(cmd='bar', help='Run bar command')
def bar_main(**kwargs):
    pass


sub_cmd = cli.add_sub_parser(cmd='sub', description='A sub command')
sub_cmd.add_option(None, '--test', help='Test mode')


@sub_cmd.command(cmd='bar', help='Run baz command')
def baz_bar_main(
        **kwargs
):
    pass


@sub_cmd.command(cmd='toto', help='Run toto command')
def toto_main(
        test: bool,
        quiet: bool,
        verbose: bool,
        foo1: int = Option(..., help='Foo arguments'),
        foo2: str = Option(..., '-f', '--foo2', help='Foo2 arguments'),
):
    pass


if __name__ == '__main__':
    cli.run()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['piou', 'piou.formatter']

package_data = \
{'': ['*']}

install_requires = \
['rich>=10.11.0,<11.0.0']

setup_kwargs = {
    'name': 'piou',
    'version': '0.1.5',
    'description': 'A CLI tool',
    'long_description': "# Piou  \n\n\n[![CircleCI](https://circleci.com/gh/Andarius/piou/tree/master.svg?style=shield)](https://app.circleci.com/pipelines/github/Andarius/piou?branch=master)\n[![Latest PyPI version](https://img.shields.io/pypi/v/piou?logo=pypi)](https://pypi.python.org/pypi/piou)\n[![Latest conda-forge version](https://img.shields.io/conda/vn/conda-forge/piou?logo=conda-forge)](https://anaconda.org/conda-forge/piou)  \n\nA CLI tool to build beautiful command-line interfaces with type validation.\n\nIt is as simple as\n\n```python\nfrom piou import Cli, Option\n\ncli = Cli(description='A CLI tool')\n\n\n@cli.command(cmd='foo',\n             help='Run foo command')\ndef foo_main(\n    foo1: int = Option(..., help='Foo arguments'),\n    foo2: str = Option(..., '-f', '--foo2', help='Foo2 arguments'),\n    foo3: str = Option(None, '-g', '--foo3', help='Foo3 arguments'),\n):\n    pass\n\nif __name__ == '__main__':\n    cli.run()\n```\nThe output will look like this: \n\n![example](https://github.com/Andarius/piou/raw/master/docs/simple-output.png)\n\n\nYou can also add sub-commands:\n\n```python\nfrom piou import Cli, Option\n\ncli = Cli(description='A CLI tool')\n\nsub_cmd = cli.add_sub_parser(cmd='sub', description='A sub command')\n\nsub_cmd.add_option(False, '-', '--dry-run', help='Run in dry run mode')\n\n\n@sub_cmd.command(cmd='foo', help='Run baz command')\ndef baz_bar_main(**kwargs):\n    pass\n\n\n@sub_cmd.command(cmd='bar', help='Run toto command')\ndef toto_main(\n        foo1: int = Option(..., help='Foo arguments'),\n        foo2: str = Option(..., '-f', '--foo2', help='Foo2 arguments'),\n):\n    pass\n\n\nif __name__ == '__main__':\n    cli.run()\n\n```\n\n![example](https://github.com/Andarius/piou/raw/master/docs/sub-cmd-output.png)\n\nHere is a more complete example: \n\n```python\nfrom piou import Cli, Option\n\ncli = Cli(description='A CLI tool')\n\ncli.add_option(None, '-q', '--quiet', help='Do not output any message')\ncli.add_option(None, '--verbose', help='Increase verbosity')\n\n\n@cli.command(cmd='foo',\n             help='Run foo command')\ndef foo_main(\n        quiet: bool,\n        verbose: bool,\n        foo1: int = Option(..., help='Foo arguments'),\n        foo2: str = Option(..., '-f', '--foo2', help='Foo2 arguments'),\n        foo3: str = Option(None, '-g', '--foo3', help='Foo3 arguments'),\n):\n    pass\n\n\n@cli.command(cmd='bar', help='Run bar command')\ndef bar_main(**kwargs):\n    pass\n\n\nsub_cmd = cli.add_sub_parser(cmd='sub', description='A sub command')\nsub_cmd.add_option(None, '--test', help='Test mode')\n\n\n@sub_cmd.command(cmd='bar', help='Run baz command')\ndef baz_bar_main(\n        **kwargs\n):\n    pass\n\n\n@sub_cmd.command(cmd='toto', help='Run toto command')\ndef toto_main(\n        test: bool,\n        quiet: bool,\n        verbose: bool,\n        foo1: int = Option(..., help='Foo arguments'),\n        foo2: str = Option(..., '-f', '--foo2', help='Foo2 arguments'),\n):\n    pass\n\n\nif __name__ == '__main__':\n    cli.run()\n```\n\nTry it by running `python -m piou.test`\n",
    'author': 'Julien Brayere',
    'author_email': 'julien.brayere@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/andarius/pioupiou',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass, field
from functools import wraps
from typing import Callable
from typing import get_type_hints, Optional, Any

from .exceptions import DuplicatedCommandError
from .utils import (
    CommandOption,
    get_default_args,
    parse_input_args,
    convert_args_to_dict
)

ParentArgs = list[tuple[str, list[CommandOption]]]


@dataclass
class Command:
    name: str
    help: Optional[str]
    fn: Callable
    options: list[CommandOption] = field(default_factory=list)

    @property
    def positional_args(self) -> list[CommandOption]:
        return [opt for opt in self.options if opt.is_positional_arg]

    @property
    def keyword_args(self) -> list[CommandOption]:
        return [opt for opt in self.options if not opt.is_positional_arg]

    def run(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

    def __post_init__(self):
        keyword_params = [x for x in self.options if not x.is_positional_arg]
        if not keyword_params:
            return

        _keyword_args = set()
        for _param in keyword_params:
            for _keyword_arg in _param.keyword_args:
                if _keyword_arg in _keyword_args:
                    raise ValueError(f'Duplicate keyword args found "{_keyword_arg}"')
                _keyword_args.add(_keyword_arg)


@dataclass
class CommandGroup:
    help: Optional[str] = None
    name: Optional[str] = None

    # _formatter: Formatter = field(init=False, default=None)
    _options: list[CommandOption] = field(init=False, default_factory=list)
    _commands: dict[str, Command] = field(init=False, default_factory=dict)
    _command_groups: dict[str, 'CommandGroup'] = field(init=False, default_factory=dict)

    @property
    def commands(self):
        return {k: self._commands.get(k) or self._command_groups[k] for k in sorted(self.command_names)}

    @property
    def options(self):
        return self._options

    def add_sub_parser(self, description: Optional[str] = None):
        cls = type(self)
        return cls(help=description)  # noqa

    def add_option(self, default: Any, *args: str, help: str = None, data_type: Any = None):
        opt = CommandOption(
            default=default,
            help=help,
            keyword_args=args,
        )
        opt.data_type = data_type
        self._options.append(opt)

    @property
    def command_names(self) -> set[str]:
        return set(self._commands.keys()) | (self._command_groups.keys())

    def add_group(self, group: 'CommandGroup'):
        if group.name is None:
            raise NotImplementedError('A group must have a name')

        if group.name in self.command_names:
            raise DuplicatedCommandError(f'Duplicated command found for {group.name!r}',
                                         group.name)
        self._command_groups[group.name] = group

    def command(self, cmd: str, help: str = None):
        if cmd in self.commands:
            raise DuplicatedCommandError(f'Duplicated command found for {cmd!r}',
                                         cmd)

        def _command(f):
            @wraps(f)
            def wrapper(*args, **kwargs):
                return f(*args, **kwargs)

            options: list[CommandOption] = []
            defaults: list[Optional[Any]] = get_default_args(f)

            for (param_name, param_type), option in zip(get_type_hints(f).items(),
                                                        defaults):
                if not isinstance(option, CommandOption):
                    continue

                option.name = param_name
                option.data_type = param_type
                options.append(option)

            self._commands[cmd] = Command(name=cmd,
                                          fn=wrapper,
                                          options=options,
                                          help=help)
            return wrapper

        return _command

    def run_with_args(self, *args, parent_args: ParentArgs = None):

        cmd, global_options, cmd_options = parse_input_args(args, self.command_names)

        command_group = self._command_groups.get(cmd) if cmd else None
        command = (command_group or self)._commands.get(cmd) if cmd else None

        # print(command)
        if command_group:
            if cmd is None:
                raise NotImplementedError('"cmd" cannot be empty')
            parent_args = [*(parent_args or [])] + [(cmd, self.options)]
            return command_group.run_with_args(*cmd_options, parent_args=parent_args)

        if set(global_options + cmd_options) & {'-h', '--help'} or \
                not command:
            raise ShowHelpError(
                group=command_group or self,
                parent_args=parent_args,
                command=command
            )

        _cmd_args_dict = convert_args_to_dict(
            cmd_options,
            command.options
        )
        _global_args_dict = convert_args_to_dict(
            global_options, self._options)

        _args_dict = {**_global_args_dict, **_cmd_args_dict}
        return command.run(**_args_dict)


class ShowHelpError(Exception):
    def __init__(self,
                 group: CommandGroup,
                 command: Command = None,
                 parent_args: ParentArgs = None):
        self.command = command
        self.group = group
        self.parent_args = parent_args



class DuplicatedCommandError(Exception):
    def __init__(self, msg: str, cmd: str):
        super().__init__(msg)
        self.cmd = cmd


class PosParamsCountError(Exception):
    def __init__(self, msg: str, expected_count: int, count: int):
        super().__init__(msg)
        self.expected_count = expected_count
        self.count = count


class ParamNotFoundError(Exception):
    def __init__(self, msg: str, key: str):
        super().__init__(msg)
        self.key = key


class CommandNotFoundError(Exception):
    def __init__(self, cmd: str):
        self.cmd = cmd
